from datetime import datetime
import json
from typing import Optional
import numpy as np
from odps import DataFrame
import pandas as pd

from tagging_index.maxcompute.sql_runner import SqlRunner
from .tag_validator import TagValidator
from tagging_index.utils.tag_tree import TagTree
from tagging_index._udf.tag_converter import TagConverter, TagStandardizer


class TagProcessor:
    def __init__(self):
        """
        process:
        1. load new config from folder should include 3 files:
            - tag_list.csv
            - prefix.csv
            - suffix.csv
            or load from json
            you can append more than one new config

        ```python
        >>> processor = TagProcessor()
        >>> processor.append_new_config_csv(tag_config_folder)
        >>> processor.append_new_config_json(tag_config_json_file)
        ```
        2. load current config from odps dim table or from local json file, if not specified, not compare with current config
        ```python
        >>> processor.load_current_config()
        >>> processor.load_local_current_config(current_config_file)
        ```
        3. validate the config and return log
        ```python
        >>> log = processor.validate_config()
        ```
        4. save the new config to odps dim table or json file. make sure all error fixed before save
        ```python
        >>> processor.save_to_json(file_name)
        >>> processor.save_to_version()
        ```
        """
        self._converter = TagConverter()
        self._standardizer = TagStandardizer()
        self._current_tags = None
        self._new_tags = []
        self.tree: TagTree = TagTree([])
        self._merge_current_config = True

    def append_new_config_csv(self, file_path: str):
        """load the raw tag config file and basic check. file_path should include 3 files:
        - tag_list.csv
        - prefix.csv
        - suffix.csv
        """
        df = self._raw_csv_to_df(file_path)
        config = self._converter.df_to_config(df)
        self._new_tags.extend(config)

    def append_new_config_json(self, file: str):
        """load the raw tag config file and basic check"""
        config = self._converter.dict_to_config(json.load(open(file)))
        self._new_tags.extend(config)

    def load_current_config(self, version: str = ""):
        """load current config from odps dim table, if version is not specified, load the latest version

        Args:
            version (str, optional): _description_. Defaults to "".
        """
        self._current_tags = self._converter.df_to_config(
            self._converter.version_to_df(version)
        )

    def load_local_current_config(self, config_file: str):
        """load current config from local config file, load_current_config and load_local_current_config are mutually exclusive

        Args:
            config_file (str): json file config
        """
        with open(config_file, "r") as file:
            current_config_dict = json.load(file)
            self._current_tags = self._converter.dict_to_config(current_config_dict)

    def _raw_csv_to_df(self, file_path: str):
        """convert local csv raw tag config to df
        Args:
            file_path (str): file_path must include and only include 3 files: tag_list.csv, prefix.csv, suffix.csv
        """
        import pandas as pd

        tag_list = pd.read_csv(filepath_or_buffer=f"{file_path}/tag_list.csv")
        # TagValidator.check_duplicated_tags(tag_list, True)
        prefix_list = pd.read_csv(f"{file_path}/prefix.csv")
        suffix_list = pd.read_csv(f"{file_path}/suffix.csv")
        tag_list = tag_list[
            [
                "tag",
                "parent_tag",
                "is_keyword",
                "data_phase",
                "category",
                "language",
            ]
        ]
        prefix_list = prefix_list[["tag", "kw", "data_phase", "category", "distance"]]
        suffix_list = suffix_list[["tag", "kw", "data_phase", "category", "distance"]]
        std = TagStandardizer()
        self._tag_list = std.apply_df(tag_list)
        self._prefix_list = std.apply_df(prefix_list)
        self._suffix_list = std.apply_df(suffix_list)
        return self._merge_prefix_suffix(tag_list, prefix_list, suffix_list)

    def reset_conifg(self):
        self._current_tags = None
        self._new_tags = []
        self.tree = TagTree([])

    def _merge_prefix_suffix(
        self, tag_df: pd.DataFrame, prefix_df: pd.DataFrame, suffix_df: pd.DataFrame
    ):
        import pandas as pd

        prefix_aggregated = (
            pd.DataFrame()
            if prefix_df.empty
            else prefix_df.groupby("tag")
            .apply(
                lambda x: x[["kw", "distance", "category", "data_phase"]].to_dict(
                    "records"
                ),
                include_groups=False,
            )
            .reset_index(name="prefix")
        )
        suffix_aggregated = (
            pd.DataFrame()
            if suffix_df.empty
            else (
                suffix_df.groupby("tag")
                .apply(
                    lambda x: x[["kw", "distance", "category", "data_phase"]].to_dict(
                        "records"
                    ),
                    include_groups=False,
                )
                .reset_index(name="suffix")
            )
        )
        # Ensure 'parent_tags' is a list of strings, not a single string
        if "parent_tag" in tag_df.columns:
            tag_df["parent_tags"] = tag_df["parent_tag"].apply(
                lambda x: [x] if pd.notna(x) else []
            )
            tag_df.drop("parent_tag", axis=1, inplace=True)
        # Merge prefix and suffix data into the tag dataframe
        if not prefix_aggregated.empty:
            tag_df = tag_df.merge(prefix_aggregated, on="tag", how="left")
        else:
            tag_df["prefix"] = None
        if not suffix_aggregated.empty:
            tag_df = tag_df.merge(suffix_aggregated, on="tag", how="left")
        else:
            tag_df["suffix"] = None
        tag_df.replace(np.nan, None, inplace=True)
        return tag_df

    def validate(self):
        """compare the raw tag config file with certain version's existing tags
        1. check duplicated tags
        2. check parent hierarchy
        3. check tag update difference
        4. check tag match overlap
        Args:
            version (str, optional): _description_. Defaults to "", use latest version.
        return dict: {
            "summary": { err_cnt:int,new_tag_cnt:int,update_tag_cnt:int },
            "dupldated_tags": [],
            "missed_parents": [],
            "circular_references": [path],
            "overlap_config": {tag:[tag...]},
            "tag_update": {field: {tag: change}
            "new_tags": {tag:parent_tag}
            },
        }
        """

        validator = TagValidator(self._new_tags, self._current_tags)
        validator.check_tags()
        validator.validate_parent_hierarchy()
        validator.check_updated_config()
        validator.check_match_overlap()
        self.has_error = (
            validator.duplicated_tags
            or validator.missed_parents
            or validator.circular_references
            or validator.overlap_config
        )
        self.validator = validator
        err_cnt = (
            len(validator.duplicated_tags)
            + len(validator.missed_parents)
            + len(validator.circular_references)
            + len(validator.overlap_config)
        )
        # count distinct tag count from validate.tag_update's value dict keys
        update_tag_cnt = len({k for v in validator.tag_update.values() for k in v})
        # new tag cnt = new_config key not in current_config key
        new_tags = {
            k: v.parent_tags[0] if v.parent_tags else ""
            for k, v in validator.new_config.items()
            if k not in validator.current_config
        }
        new_tag_cnt = len(new_tags)
        return {
            "summary": {
                "err_cnt": err_cnt,
                "new_tag_cnt": new_tag_cnt,
                "update_tag_cnt": update_tag_cnt,
            },
            "dupldated_tags": validator.duplicated_tags,
            "missed_parents": validator.missed_parents,
            "circular_references": validator.circular_references,
            "overlap_config": validator.overlap_config,
            "tag_update": validator.tag_update,
            "new_tags": new_tags,
        }

    def save_to_json(self, output_file: str, merge_current_config: bool = True):
        self._converter.config_to_json(
            list(self.validator.get_valid_config(merge_current_config).values()),
            output_file,
        )

    def save_to_version(self, merge_current_config: bool = True):
        tmp_table = self.save_to_temp_table(merge_current_config)
        self._insert_partition(tmp_table, "dim_digitalization_tag")
        result = f"version:{self._tag_version}"
        return result

    def save_to_temp_table(self, merge_current_config: bool = True):
        merged_configs = self.validator.get_valid_config(merge_current_config)
        df = self._converter.config_to_df(list(merged_configs.values()))
        tmp_table = self._persist_tmp_table(self._standardizer.df_serialize(df))
        return tmp_table

    def show_tree(
        self, root_tag: Optional[str] = None, levels=0, merge_current_config=True
    ):
        """print the tag tree

        Args:
            root_tag (Optional[str], optional): print the sub tree of root_tag if specified. Defaults to None.
            levels (int, optional): how many levels to print from root_tag. Defaults to 0.
            merge_current_config (bool, optional): combine current config with new config and print. Defaults to True.
        """
        self._init_tree(merge_current_config)
        self.tree.show_level(root_tag, levels)

    def _init_tree(self, merge_current_config=True) -> None:
        if self._merge_current_config != merge_current_config:
            self.tree = TagTree([])
        if self.tree.size() == 0:
            config = self.validator.get_valid_config(merge_current_config)
            self.tree = TagTree(list(config.values()))

    # def find_tag(self, tag: str, merge_current_config=True):
    #     """find tag and show the tag path"""
    #     self._init_tree(merge_current_config)
    #     self.tree.show_level(tag, levels=1)

    def _persist_tmp_table(self, df: pd.DataFrame) -> str:
        now = datetime.now()
        # 格式化日期和时间
        formatted_date = now.strftime("%Y%m%d")
        formatted_time = now.strftime("%H%M%S").zfill(6)
        # 合并日期和时间
        tag_version = formatted_date + formatted_time
        df["tag_version"] = tag_version
        tmp_table_name = f"tmp_tag_{tag_version}"
        # 写表指定字符串
        as_type = {
            col: "string" if dtype != bool else dtype.name
            for col, dtype in df.dtypes.items()
        }
        odps_df = DataFrame(df, as_type=as_type)
        odps_df["tag"] = odps_df["tag"].replace(r"[\n\r\t]", "")
        # 写入分区表
        odps_df.persist(tmp_table_name, lifecycle=1)
        self._tag_version = tag_version
        return tmp_table_name

    def _insert_partition(self, tmp_table_name: str, target_table: str):
        sql = f"""
        INSERT OVERWRITE TABLE {target_table} PARTITION (tag_version)
            SELECT
            tag,
            cast(tag_level as int),
            from_json(prefix,"ARRAY<STRUCT<kw:STRING,distance:INT,category:STRING,data_phase:STRING>>"),
            from_json(suffix,"ARRAY<STRUCT<kw:STRING,distance:INT,category:STRING,data_phase:STRING>>"),
            cast(is_keyword as BOOLEAN),
            language,
            category,
            data_cycle,
            data_phase,
            from_json(parent_tags,"ARRAY<STRING>"),
            tag_version
            FROM {tmp_table_name} 
            ; 
            """
        # 同步执行
        SqlRunner(sql).run()



from typing import Tuple, Type
import pandas as pd
from odps import ODPS

from tagging_index.maxcompute.sql_runner import SqlRunner
from .common_index_generator import CommonIndexGenerator, IndexSqlTemplate


class IndexIntersection:
    def __init__(self, index_generator: CommonIndexGenerator):
        self.index_generator = index_generator

    def _generate_sql(self, index_codes: Tuple[str, str], return_result=False):
        index1 = self.index_generator.get_index_sql_template(index_codes[0])
        index2 = self.index_generator.get_index_sql_template(index_codes[1])
        intersection_sql = f"""
        (select a.* from {index1.sql_source_tag_data} a left semi join {index2.sql_source_tag_data} b
        on a.entity_id=b.entity_id)
        """
        index_x_template = IndexSqlTemplate(
            index1.sql_comp_map,
            intersection_sql,
            index1.sql_year_sequence,
            index1.index_version,
            index1.comp_type,
            f"X_{index_codes[0]}_{index_codes[1]}",
            f"X_{index1.index_name}_{index2.index_name}",
            self.index_generator.DWD_TABLE,
        )
        sql_template = (
            index1.INDEX_TEMPLATE_SELECT
            if return_result
            else index1.INDEX_TEMPLATE_INSERT
        )
        sql = sql_template.format(**index_x_template.__dict__)
        return sql

    def get_index_data(self, index_codes: Tuple[str, str]) -> pd.DataFrame:
        """return specified index data in dataframe rather than insert into table

        Args:
            index_code (str)

        Returns:
            pd.DataFrame:
        """
        sql = self._generate_sql(index_codes, return_result=True)
        df = SqlRunner(sql).to_pandas()
        return df



from dataclasses import dataclass, field
from math import e
from typing import Dict, List, Optional, Type
from tagging_index.index_generator.common_index_generator import (
    CommonIndexGenerator,
    TIndexGenerator,
)
from tagging_index.maxcompute.sql_runner import SqlRunner


@dataclass
class VariableMap:
    code: str
    column_name: Optional[str] = field(default=None)
    """if not provide, use code as column_name"""


@dataclass
class VariableMapOther:
    source_data: str
    """ the table or sql for source data, sql should bracket off """
    source_var_col: str
    var_col_name: Optional[str] = field(default=None)
    """if not provide, use source_var_col as var_col_name"""
    dim_comp_id: str = "comp_id"
    col_comp_id: str = "comp_id"
    col_year: Optional[str] = field(default=None)
    """ none for non-time sequence variable """


class PanelDataGenerator:

    def __init__(self):
        self.index_version = ""
        self.index_dwd_table = "dwd_digital_index"
        self.matrix_dwd_table = "dwd_cn_lst_com_measure"
        self.dim_comp_table = """
        ( 
                SELECT full_stock_id as comp_id
                ,stock_id
                ,nvl(abbr_cn,company_name_cn) company_abbr
                ,industry_level1 as industry_name
                FROM dim_china_listed_company 
                WHERE pt = MAX_PT('dim_china_listed_company')
        )
        """
        self.comp_type = "lst_com"
        self._index_var: List[VariableMap] = []
        # total count of source
        self._index_base: List[VariableMap] = []
        self._matrix_var: List[VariableMap] = []
        self.start_year = 2016
        self.end_year = 2020
        self._other_var: Dict[str, VariableMapOther] = {}
        self.comp_ids = []
        """empty for all comp_ids, default is all comp_ids"""

    def _get_index_cols(self):
        return (
            "," + "\n,".join([x.column_name or x.code for x in self._index_var])
            if self._index_var
            else ""
        )

    def _get_index_fields(self):
        return "\n,".join(
            f"'{x.code}' as {x.column_name or x.code}" for x in self._index_var
        )

    def _get_index_codes(self):
        return ",".join([f"'{x.code}'" for x in self._index_var])

    def _get_base_index_cols(self):
        return (
            "," + "\n,".join([x.column_name or x.code for x in self._index_base])
            if self._index_base
            else ""
        )

    def _get_base_index_fields(self):
        return "\n,".join(
            f"'{x.code}-total' as {x.column_name or x.code}" for x in self._index_base
        )

    def _get_matrix_fields(self):
        return "\n,".join(
            f"'{x.code}' as {x.column_name or x.code}" for x in self._matrix_var
        )

    def _get_matrix_cols(self):
        return (
            "," + ",".join([x.column_name or x.code for x in self._matrix_var])
            if self._matrix_var
            else ""
        )

    def _get_matrix_codes(self):
        return ",".join([f"'{x.code}'" for x in self._matrix_var])

    def add_index(self, code: str, column_name: Optional[str] = None):
        """add index data to panel

        Args:
            code (str): index code to add to panel's variable
            column_name (Optional[str], optional): if not provide, use code as column_name. Defaults to None.
        """
        self._index_var.append(VariableMap(code, column_name or code))

    def add_source_base(
        self,
        source_class: Type[CommonIndexGenerator],
        column_name: Optional[str] = None,
    ):
        self._index_base.append(
            VariableMap(
                source_class.get_source_name(),
                column_name or source_class.get_source_name(),
            )
        )

    def add_matrix(self, code: str, column_name: Optional[str] = None):
        self._matrix_var.append(VariableMap(code, column_name or code))

    def _get_com_year_squence_sql(self):
        return f"""(select /*+ MAPJOIN(y) */ * from 
        {self.dim_comp_table} a join {self._get_year_sequence()} y
        where {self._get_comp_condition()}
        )"""

    def _get_index_sql(self):
        if self._index_var:
            sql = f"""(SELECT * from
            ( select comp_id,eff_year,index_code,index_value from {self.index_dwd_table}  
                where index_version = '{self.index_version}'  
                and comp_type = '{self.comp_type}'
                and eff_year in {self._get_year_sequence()} and {self._get_comp_condition()}
                and index_code in ({self._get_index_codes()})
                )
            pivot (max(index_value) for index_code IN ({self._get_index_fields()}))
            )
            """
            return sql
        else:
            return ""

    def _get_base_index_sql(self):
        if self._index_base:
            sql = f"""(SELECT * from
        ( select comp_id,eff_year,index_version,index_value from {self.index_dwd_table}  
            where comp_type = '{self.comp_type}'
            and eff_year in {self._get_year_sequence()} and {self._get_comp_condition()}
            and index_code ='ttl_cnt'
            )
        pivot (max(index_value) for index_version IN ({self._get_base_index_fields()}))
        )
        """
            return sql
        else:
            return ""

    def _get_matrix_sql(self):
        if not self._matrix_var:
            return ""
        sql = f"""(SELECT * from 
            (select full_stock_id as comp_id,eff_year,source_code,measure_value from {self.matrix_dwd_table} 
             where eff_year in {self._get_year_sequence()} 
             and {self._get_comp_condition("full_stock_id")}
             and source_code in ({self._get_matrix_codes()})
            )
            pivot (max(measure_value) for source_code IN ({self._get_matrix_fields()}))
            )
        """
        return sql

    def _get_year_sequence(self):
        return f"(SELECT EXPLODE(SEQUENCE({self.start_year},{self.end_year})) as (eff_year))"

    def _get_comp_condition(self, comp_id_col="comp_id"):
        return (
            f""" {comp_id_col} in ({','.join([f"'{x}'" for x in self.comp_ids])})"""
            if self.comp_ids
            else "1=1"
        )

    def get_panel_sql(self):
        if self.index_version == "":
            raise Exception("index_version not set")
        if self.comp_type == "":
            raise Exception("comp_type not set")
        idx_sql = self._get_index_sql()
        base_idx_sql = self._get_base_index_sql()
        idx_join_sql = (
            f"""\nleft join {idx_sql} b 
        on a.comp_id =b.comp_id and a.eff_year=b.eff_year"""
            if idx_sql
            else ""
        )
        base_join_sql = (
            f"""\nleft join {base_idx_sql} d
        on a.comp_id =d.comp_id and a.eff_year=d.eff_year"""
            if base_idx_sql
            else ""
        )
        matrix_sql = self._get_matrix_sql()
        matrix_join_sql = (
            f"""\nleft join {matrix_sql} c  
        on a.comp_id =c.comp_id and a.eff_year=c.eff_year"""
            if matrix_sql
            else ""
        )

        other_vars = self._get_other_var_sql()
        other_vars_join_sql = "\n".join(
            [""]
            + [
                f"""left join {x} as {v} on {v}.comp_id = a.comp_id {f'and {v}.eff_year = a.eff_year' 
                if self._other_var[v].col_year else ''}"""
                for v, x in other_vars.items()
            ]
        )
        other_vars_cols = (
            "," + "\n,".join([f"{v}.{v}" for v, _ in other_vars.items()])
            if other_vars
            else ""
        )
        if (
            not idx_join_sql
            and not base_join_sql
            and not matrix_join_sql
            and not other_vars_cols
        ):
            raise Exception("please add at least one variable!")
        sql = f""" select a.comp_id,a.eff_year,a.company_abbr,a.industry_name
        {self._get_matrix_cols()}
        {self._get_base_index_cols()}
        {self._get_index_cols()}
        {other_vars_cols}
        from {self._get_com_year_squence_sql()} a
        {idx_join_sql}{base_join_sql} {matrix_join_sql}{other_vars_join_sql}
        ;
        """
        return sql

    def add_other_var(self, var_map: VariableMapOther):
        if self._other_var.get(var_map.var_col_name or var_map.source_var_col):
            raise Exception(
                f"other var {var_map.var_col_name or var_map.source_var_col} already exists"
            )
        self._other_var[var_map.var_col_name or var_map.source_var_col] = var_map

    def _get_other_var_sql(self):
        other_var_sql: Dict[str, str] = {}
        for x, v in self._other_var.items():
            eff_year = f" a.{v.col_year} as eff_year," if v.col_year else ""
            where = (
                f" where  {v.col_year} in {self._get_year_sequence()}"
                if v.col_year
                else ""
            )
            other_var_sql[x] = (
                f"""(
            SELECT  b.comp_id,{eff_year}a.{x}
            from (
                select {v.col_comp_id},{v.col_year+',' if v.col_year else ""}{v.source_var_col} as {x} from {v.source_data} {where}
            ) a join {self.dim_comp_table} b on {v.col_comp_id} = b.{v.dim_comp_id})
            """
            )
        return other_var_sql

    def get_result_df(self):
        sql = self.get_panel_sql()
        result = SqlRunner(sql).to_pandas()
        return result

    def save_to_csv(self, path):
        df = self.get_result_df()
        df.to_csv(path, index=False)



"""generate index in maxcompute for talent, demand, innovation etc, maxcompute should include required datasource
available index:
    TalentIndexGenerator: data from linkedin positions
    DemandIndexGenerator: data from job posting
    InnovationIndexGenerator: data from patent data 
"""

from .talent_index_generator import TalentIndexGenerator
from .demand_index_generator import DemandIndexGenerator
from .innovation_index_generator import InnovationIndexGenerator
from .index_intersection import IndexIntersection
from .common_index_generator import TIndexGenerator



from collections import Counter

import re
from typing import List, Dict, Optional
from dataclasses import dataclass, field

from .tag_converter import (
    PreSuffixConfig,
    TagConfig,
    TagStandardizer,
)


@dataclass
class TagMatch:
    prefix: Optional[str]
    suffix: Optional[str]
    p_distance: int
    s_distance: int
    tag: str
    language: str
    tag_level: int
    match_length: int
    category: Optional[str]
    data_phase: Optional[str]
    data_cycle: Optional[str]
    parent_tags: List[str]
    basic_match: str
    pattern: Optional[re.Pattern] = field(default=None)


@dataclass
class MatchResult:
    prefix: Optional[str]
    suffix: Optional[str]
    tag: str
    category: Optional[str]
    data_phase: Optional[str]
    data_cycle: Optional[str]
    matched: str


class TagMatcher:
    def __init__(self):
        self.tag_config: Dict[str, TagConfig] = {}
        self.tag_match_list: List[TagMatch] = []
        self.standardizer = TagStandardizer()

    def load_tags(self, tag_data: List[TagConfig]):
        tc_list = self.standardizer.apply_config(tag_data)
        tc_dict = {i.tag: i for i in tc_list}
        # Inherit parent tag settings by sorting based on tag level
        sorted_configs = sorted(tc_dict.values(), key=lambda tm: tm.tag_level)
        for p in sorted_configs:
            for parent_tag in p.parent_tags:
                if parent_tag in tc_dict:
                    self.merge_parent_configs(p, tc_dict[parent_tag])
        self.tag_config = tc_dict
        self.tag_match_list = self.parse_tag_match()

    def parse_tag_match(self) -> List[TagMatch]:
        tag_list: List[TagMatch] = []
        for p in self.tag_config.values():
            if p.is_keyword:
                basic_match = f" {p.tag} " if p.language == "EN" else p.tag
                match_length = len(basic_match.strip())
                tag_list.append(
                    TagMatch(
                        None,
                        None,
                        0,
                        0,
                        p.tag,
                        p.language,
                        p.tag_level,
                        match_length,
                        p.category,
                        p.data_phase,
                        p.data_cycle,
                        p.parent_tags,
                        basic_match,
                    )
                )
            for pre_config in p.prefix or []:
                tag_list.extend(self.parse_tag_pre_suffix(p, pre_config, True))
            for suf_config in p.suffix or []:
                tag_list.extend(self.parse_tag_pre_suffix(p, suf_config, False))

        # Sort by tag level and presence of prefix/suffix
        sorted_tag_list = sorted(
            tag_list,
            key=lambda tm: (
                -tm.tag_level,
                (-1 if tm.prefix else 0) + (-1 if tm.suffix else 0),
            ),
        )
        return sorted_tag_list

    def parse_tag_pre_suffix(
        self, tc: TagConfig, config: PreSuffixConfig, is_pre: bool
    ) -> List[TagMatch]:
        """
        Parses a single prefix or suffix configuration and returns a list of TagMatch objects.

        Args:
            param (TagConfig): The main tag configuration.
            config (PreSuffixConfig): The prefix or suffix configuration to be parsed.
            is_pre (bool): Whether the configuration is for a prefix.

        Returns:
            List[TagMatch]: A list of TagMatch objects representing the parsed configuration.
        """

        def fix_pattern_cn(prefix, distance, suffix):
            pattern = r"{}[^\W\s_]{{{}}}{}".format(
                prefix, "{},{}".format(0, distance), suffix
            )
            return re.compile(pattern)

        def fix_pattern_en(prefix, distance, suffix):
            # 构造正则表达式，允许0-distance个单词（由[a-zA-Z]*\s*表示）在prefix和keyword之间
            pattern = r"(?i)\b{}\b(?:\s*[a-zA-Z.-]*\s*){{{}}}\b{}".format(
                prefix, "{},{}".format(0, distance), suffix
            )
            # print(pattern)
            return re.compile(pattern)

        # 模式匹配前缀和后缀, 根据当前配置是否pre决定
        pattern_pre = config.kw if is_pre else tc.tag
        pattern_suf = tc.tag if is_pre else config.kw
        # 根据中英文决定匹配方式
        basic_pattern = " {} {} " if tc.language == "EN" else "{}{}"
        fix_pattern = fix_pattern_cn if tc.language == "CN" else fix_pattern_en
        basic_match = None
        pattern = None
        # 如果距离=0 则用简单匹配
        basic_match = basic_pattern.format(pattern_pre, pattern_suf)
        if config.distance > 0:
            pattern = fix_pattern(pattern_pre, config.distance, pattern_suf)
        match_len = len(basic_match.strip())
        match = TagMatch(
            config.kw if is_pre else None,
            None if is_pre else config.kw,
            config.distance if is_pre else 0,
            0 if is_pre else config.distance,
            tc.tag,
            tc.language,
            tc.tag_level,
            match_len,
            config.category or tc.category,
            config.data_phase or tc.data_phase,
            tc.data_cycle,
            tc.parent_tags,
            basic_match,
            pattern,
        )
        return [match]

    def merge_parent_configs(self, current: TagConfig, parent: TagConfig) -> None:
        """
        Merges the prefix and suffix configurations from the parent TagConfig into the current one.

        Args:
            current (TagConfig): The child TagConfig to update.
            parent (TagConfig): The parent TagConfig providing additional configurations.
        """
        for pre_config in parent.prefix:
            if pre_config.kw not in [c.kw for c in current.prefix]:
                current.prefix.append(pre_config)
        for suf_config in parent.suffix:
            if suf_config.kw not in [c.kw for c in current.suffix]:
                current.suffix.append(suf_config)

    def get_tag_range(self, p_tag_range: List[str]) -> List[TagMatch]:
        p_tag_range = [x.lower() for x in p_tag_range]
        tag_configs: Dict[str, TagConfig] = {}

        # 一级标签
        for tag, config in self.tag_config.items():
            if tag in p_tag_range:
                tag_configs[tag] = config

        # 二级及更高级别的标签
        processed_tags = set(tag_configs.keys())
        while True:
            new_tags_added = False
            for tag, config in self.tag_config.items():
                if tag not in processed_tags and any(
                    parent in processed_tags for parent in config.parent_tags
                ):
                    tag_configs[tag] = config
                    processed_tags.add(tag)
                    new_tags_added = True

            if not new_tags_added:
                break
        result: List[TagMatch] = [
            x for x in self.tag_match_list if x.tag in tag_configs
        ]
        return result

    def match_content(
        self, content: str, tag_range: List[str] = []
    ) -> List[MatchResult]:

        match_results: List[MatchResult] = []
        if content:
            content = content.lower().strip()
            content_len = len(content)
            # 需要copy, 因为tag_list会改变
            tag_list = (
                self.get_tag_range(tag_range)
                if tag_range
                else self.tag_match_list.copy()
            )

            for x in tag_list:
                # 仅当content 长度大于tag+ifx长度时才有可能匹配到
                if content_len < x.match_length:
                    continue
                # 简单匹配
                matched = None
                # add content's space for EN tag match
                content_for_match = f" {content} " if x.language == "EN" else content
                # 不论是否有pattern,先简单匹配
                # if not x.pattern:
                matched = x.basic_match if x.basic_match in content_for_match else None
                # 仅当简单匹配没有匹配到时才进行正则匹配
                # else:
                if not matched and x.pattern:
                    re_match = re.search(x.pattern, content_for_match)
                    matched = re_match.group() if re_match else None
                if matched:
                    match_results.append(
                        MatchResult(
                            x.prefix,
                            x.suffix,
                            x.tag,
                            x.category,
                            x.data_phase,
                            x.data_cycle,
                            matched.strip(),
                        )
                    )
                    # 如果匹配到，则去除后续的父标签及无前缀、后缀标签
                    if x.prefix or x.suffix:
                        for t in reversed(tag_list[tag_list.index(x) + 1 :]):
                            if t.prefix is None and t.suffix is None and t.tag == x.tag:
                                tag_list.remove(t)
                    if x.parent_tags and len(x.parent_tags):
                        for p_tag in x.parent_tags:
                            for t in reversed(tag_list[tag_list.index(x) + 1 :]):
                                if (
                                    t.prefix is None
                                    and t.suffix is None
                                    and t.tag == p_tag
                                ):
                                    tag_list.remove(t)
        return match_results

    def match_result(
        self,
        content: str,
        return_type: int = 0,
        tag_range: List[str] = [],
        tag_verion="DEV",
        udf_version="DEV",
    ) -> List[str]:
        match_results = self.match_content(content, tag_range)
        # for debug use
        if return_type == -1:
            tag_list = (
                self.get_tag_range(tag_range)
                if tag_range
                else self.tag_match_list.copy()
            )
            return [
                "{}_{}_{}_{}.{}".format(
                    x.tag,
                    x.prefix,
                    x.suffix,
                    tag_verion,
                    udf_version,
                )
                for x in tag_list
            ]
        return_results = []
        for r in match_results:
            if return_type == 0:
                formated_result = "{}_{}_{}_{}_{}_{}.{}".format(
                    r.matched,
                    r.tag,
                    r.category,
                    r.data_phase,
                    r.data_cycle,
                    tag_verion,
                    udf_version,
                )
            elif return_type == 1:  # 标志出prefix和tag匹配结果
                formated_result = r.matched
            elif return_type == 2:  # 输出匹配的纯文本
                formated_result = r.tag
            # return category only
            elif return_type == 3:
                formated_result = r.category
            # return data_phase only
            elif return_type == 4:
                formated_result = r.data_phase
            # return category & data cycle
            elif return_type == 5:
                formated_result = "{}_{}_{}".format(
                    r.category, r.data_phase, r.data_cycle
                )
            else:
                formated_result = r.tag
            return_results.append(formated_result)
        counter_results = Counter(return_results)
        return (
            [
                "{}@{}".format(element, count)
                for element, count in counter_results.items()
            ]
            if return_type > 1
            else return_results
        )



from typing import List, Optional
import pandas as pd

from tagging_index._udf.tag_converter import (
    TagStandardizer,
    PreSuffixConfig,
    TagConfig,
    TagConverter,
)
from tagging_index._udf.tag_matcher import TagMatcher


class TagValidator:
    def __init__(
        self,
        new_conf_list: List[TagConfig],
        cur_conf_list: Optional[List[TagConfig]] = None,
    ):
        """validate new config with current config if provide

        Args:
            new_config (List[TagConfig]): new tag config
            current_config (Optional[List[TagConfig]], optional): the current tag config. Defaults to None. if not provide, will use new_config as current_config
        """
        self.stanadardizer = TagConverter()
        self.new_config_list = new_conf_list
        self.current_config_list = cur_conf_list or new_conf_list
        self.has_current_config = cur_conf_list is not None
        # convert list to dict with tag as key, last win if dup.
        self.new_config = {n.tag: n for n in self.new_config_list}
        self.current_config = {n.tag: n for n in self.current_config_list}
        self.updated_config = {}

    @staticmethod
    def check_duplicated_tags(df_tags: pd.DataFrame, raise_err=False):
        if df_tags.empty:
            return []
        if "tag" not in df_tags.columns:
            raise ValueError("DataFrame must contain a 'tag' column")
        # check column tag if duplicated and find all duplicates
        # standardize tag name before check
        df_tags["tag"] = df_tags["tag"].apply(lambda x: x.strip().lower())
        duplicates = df_tags.duplicated(subset="tag")
        dup_tags = list(set(df_tags[duplicates]["tag"]))
        if duplicates.any() and raise_err:
            raise ValueError(f"duplicated tag(s) found: {','.join(dup_tags)}")
        return dup_tags

    def check_tags(self):
        """check new config and current config's tag, return duplicated tags, missed parent tags

        Raises:
            ValueError: for current config will raise err directly

        Returns:
            (list[str],list[str]): (duplicated tags,missed parents) for new config
        """
        if self.has_current_config:
            dup_tags = self.check_duplicated_tags(
                pd.DataFrame(self.current_config_list)
            )
            if len(dup_tags) > 0:
                raise ValueError(
                    f"duplicated tag(s) found in current config: {','.join(dup_tags)}"
                )
        duplicated_tags = self.check_duplicated_tags(pd.DataFrame(self.new_config_list))
        missed_parents = self.build_tag_hierarchy()
        self.duplicated_tags = duplicated_tags
        # print(self.tag_hierarchy, self.current_tag_hierarchy)
        return (duplicated_tags, missed_parents)

    # compare prefix and suffix with serialized json
    @staticmethod
    def compare_field(new, cur):
        new = TagStandardizer.serialize(new)
        cur = TagStandardizer.serialize(cur)
        # only compare when new is not None
        if new and (new != cur):
            return f"{cur}->{new}"

    def check_updated_config(self):
        """compare the new tag config with current tag config"""
        field_list = [
            "is_keyword",
            "data_phase",
            "category",
            "data_cycle",
            "language",
        ]
        # use field_list as key for updated_config dict
        updated_config = {k: {} for k in field_list}
        updated_config["prefix"] = {}
        updated_config["suffix"] = {}

        if self.has_current_config:
            for tag, new_conf in self.new_config.items():
                if tag in self.current_config:
                    cur_conf = self.current_config[tag]
                    self.updated_config[tag] = new_conf
                    # Check for update
                    for field in field_list:
                        u = self.compare_field(
                            new_conf.__dict__[field], cur_conf.__dict__[field]
                        )
                        if u:
                            updated_config[field][tag] = u

                    # check prefix and suffix for current tag
                    updated_prefix, merged_prefix = self.check_updated_fix(
                        tag, new_conf.prefix, cur_conf.prefix
                    )
                    if updated_prefix:
                        updated_config["prefix"].update(updated_prefix)
                        # merge fix at same tag
                        self.updated_config[tag].prefix = list(merged_prefix.values())
                    updated_suffix, merged_suffix = self.check_updated_fix(
                        tag, new_conf.suffix, cur_conf.suffix
                    )
                    if updated_suffix:
                        updated_config["suffix"].update(updated_suffix)
                        self.updated_config[tag].suffix = list(merged_suffix.values())

        # check parent tags, only the first element
        updated_config["parent_tag"] = self.check_updated_parents()
        # remove empty values
        updated_config = {k: v for k, v in updated_config.items() if v}
        self.tag_update = updated_config
        return updated_config

    def check_updated_fix(
        self, tag: str, new_fix: List[PreSuffixConfig], cur_fix: List[PreSuffixConfig]
    ):
        """check prefix or suffix update, new add or config change, TODO remove kw
        Returns:
            {tag:updated_fix}
        """
        updated_fix = {"update": {}, "new": [], "remove": []}
        merged_fix = {p.kw: p for p in cur_fix}
        field_list = ["category", "distance", "data_phase"]
        new_fix_dict = {x.kw: x for x in new_fix}
        cur_fix_dict = {x.kw: x for x in cur_fix}
        if self.has_current_config and new_fix:
            for kw, new_conf in new_fix_dict.items():
                if kw in cur_fix_dict:
                    cur_conf = cur_fix_dict[kw]
                    # Check for update
                    updated = []
                    for field in field_list:
                        u = self.compare_field(
                            new_conf.__dict__[field], cur_conf.__dict__[field]
                        )
                        if u:
                            updated.append(u)
                            # update merged_fix
                            merged_fix[kw].__dict__[field] = new_conf.__dict__[field]
                    if updated:
                        updated_fix["update"][kw] = updated
                else:
                    updated_fix["new"].append(new_conf.__dict__)
                    # append merged_fix
                    merged_fix[kw] = new_conf
        updated_fix = {k: v for k, v in updated_fix.items() if v}
        return {tag: updated_fix} if updated_fix else {}, merged_fix

    def check_updated_parents(self):
        updated_parents = {}
        if self.has_current_config:
            for tag in self.tag_hierarchy:
                if tag in self.current_tag_hierarchy:
                    if self.current_tag_hierarchy[tag] != self.tag_hierarchy[tag]:
                        updated_parents[tag] = (
                            f"{self.current_tag_hierarchy[tag]}->{self.tag_hierarchy[tag]}"
                        )
        self.updated_parents = updated_parents
        return updated_parents

    def build_tag_hierarchy(self):
        """build hierarchy for new config, refer to current config if exists
            initialize self's tag_hierarchy,current_tag_hierarchy,missed_parents
        Returns:
            list[str]: missed_parents
        """
        missed_parents = set()
        # based on new config and current config to build new tag's hierarchy
        tag_hierarchy = self.get_tag_parent_pair(self.new_config_list)
        current_tag_hierarchy = self.get_tag_parent_pair(self.current_config_list)
        # check if all parent tag is in hierarchy, if not check in current config's tag
        for parent_tag in list(tag_hierarchy.values()):
            if parent_tag and parent_tag not in tag_hierarchy:
                if parent_tag in current_tag_hierarchy:
                    tag_hierarchy[parent_tag] = current_tag_hierarchy[parent_tag]
                else:
                    # record error str, return all errors
                    missed_parents.add(parent_tag)
        self.missed_parents = list(missed_parents)
        self.tag_hierarchy = tag_hierarchy
        self.current_tag_hierarchy = current_tag_hierarchy
        return self.missed_parents

    @staticmethod
    def get_tag_parent_pair(config: List[TagConfig]):
        hierarchy = {}
        for tag_config in config:
            hierarchy[tag_config.tag] = (
                tag_config.parent_tags[0] if tag_config.parent_tags else ""
            )
        return hierarchy

    def validate_parent_hierarchy(self):
        """append parent full path and level and Check for circular dependencies in tag parent-child relationships."""
        # check cycle in tag hierarchy
        circular_references = set()
        circular_tags = []
        for tag, config in self.new_config.items():
            p = config.parent_tags[0] if config.parent_tags else ""
            # include tag self to check for cycle
            tag_path = [tag]
            while p:
                tag_path.append(p)
                p = self.tag_hierarchy[p] if p in self.tag_hierarchy else ""
                if p in tag_path:
                    # already exist circular Parent tag
                    if p not in circular_tags:
                        circular_tags.append(p)
                        short_path = [t for t in tag_path[tag_path.index(p) :]]
                        short_path.append(p)
                        circular_references.add("->".join(short_path))
                    break
            # remove tag itself
            self.new_config[tag].parent_tags = tag_path[1:]
            self.new_config[tag].tag_level = len(tag_path)
        self.circular_references = list(circular_references)
        return self.circular_references

    def get_valid_config(self, merge_current_config: bool = True):

        merged_configs = self.new_config.copy()

        if merge_current_config and self.has_current_config:
            for tag in merged_configs:
                if tag in self.updated_config:
                    merged_configs[tag] = self.updated_config[tag]
            # new config'tag will override current config'tag
            append_current = {
                k: v for k, v in self.current_config.items() if k not in self.new_config
            }
            merged_configs.update(append_current)
        return merged_configs

    def check_match_overlap(self):
        self.matcher = TagMatcher()
        overlap_config = {}
        valid_config = self.get_valid_config()
        self.matcher.load_tags(list(valid_config.values()))

        for tag, conf in valid_config.items():
            if conf.is_keyword:
                match_results = self.matcher.match_content(tag)
                if len(match_results) > 1:
                    # 排除tag自身
                    for x in match_results:
                        if x.tag != tag:
                            if x.tag not in overlap_config:
                                overlap_config[x.tag] = set()
                            overlap_config[x.tag].add(tag)
        # change set to list
        for tag in overlap_config:
            overlap_config[tag] = list(overlap_config[tag])
        self.overlap_config = overlap_config
        return overlap_config



from typing import Dict, List, Type
from dataclasses import dataclass
from tagging_index.maxcompute import maxcompute_setup
from tagging_index.index_generator import TIndexGenerator
from tagging_index.maxcompute.sql_runner import SqlRunner


class DashboardAdsGenerator:
    @dataclass
    class AdsColumns:
        comp_id: str = "comp_id"
        company_abbr: str = "company_abbr"
        industry_name: str = "industry_name"
        eff_year: str = "eff_year"
        index_version: str = "index_version"
        ttl_cnt: str = "ttl_cnt"
        it_cnt: str = "it_cnt"
        source_name: str = "source_type"
        index_name: str = "index_type"
        index_value: str = "index_value_cnt"

    def __init__(self, source_cls: List[Type[TIndexGenerator]]):
        self.source_cls = source_cls
        self.comp_type = "lst_com"
        self.index_version = ""
        self._it_total_code = "IT_total"
        self.index_codes = []
        self.ads_columns = self.AdsColumns()

        self.start_year = 2000
        self.end_year = 2024
        self.comp_list = []

    def _generate_ttl_sql(self) -> Dict[str, str]:
        ttl_sql = {}
        comp_list = (
            ",".join([f'"{x}"' for x in self.comp_list]) if self.comp_list else ""
        )
        comp_cond = (
            f"comp_id in ({comp_list})" if self.comp_list else "comp_id is not null"
        )
        for c in self.source_cls:
            ttl_sql[c.get_source_name()] = (
                f"""  (
                select comp_id,index_value,eff_year 
                from {c.DWD_TABLE} where index_version = "{c.get_source_name()}-total" 
                and comp_type = "{self.comp_type}"
                and index_code = "ttl_cnt"
                and eff_year >= {self.start_year} and eff_year <= {self.end_year}
                and {comp_cond}
                )
                """
            )
        return ttl_sql

    def _get_index_ver_cond(self, cls: Type[TIndexGenerator]) -> str:
        if self.index_version:
            return f"'{self.index_version}'"
        else:
            return f"(select max(index_version) from {cls.DWD_TABLE} where index_version not like '%-total')"

    # special index for IT total
    def _generate_index_it_ttl_sql(self) -> Dict[str, str]:
        index_sql = {}
        for c in self.source_cls:
            index_sql[c.get_source_name()] = (
                f"""  (
            select comp_id,index_value,eff_year 
            from {c.DWD_TABLE} where index_version = {self._get_index_ver_cond(c)}
            and comp_type = "{self.comp_type}" and index_code = '{self._it_total_code}'
            and eff_year >= {self.start_year} and eff_year <= {self.end_year}
            )
            """
            )
        return index_sql

    def _generate_index_sql(self) -> Dict[str, str]:
        index_sql = {}

        for c in self.source_cls:
            index_code_list = ",".join(
                f"'{x}_{c.get_index_suffix()}'" for x in self.index_codes
            )
            index_sql[c.get_source_name()] = (
                f""" (
                select comp_id,index_code,index_name,index_value,eff_year,index_version
                from {c.DWD_TABLE} where index_version = {self._get_index_ver_cond(c)}
                and comp_type = "{self.comp_type}"
                and index_code in ( {index_code_list} )
                and eff_year >= {self.start_year} and eff_year <= {self.end_year}
                )
                """
            )
        return index_sql

    _dim_com = """
        ( 
                SELECT full_stock_id as comp_id
                ,nvl(abbr_cn,company_name_cn) company_abbr
                ,industry_level1 as industry_name
                FROM dim_china_listed_company 
                WHERE pt = MAX_PT('dim_china_listed_company')
        )
    """

    def get_dim_index_code(self, index_data_sql: str):
        sql = f"""(select distinct index_code,index_name,index_version from {index_data_sql})"""
        return sql

    def _generate_dashboard_sql(self, test=False):
        if self.index_codes == []:
            raise Exception("Please set index_code first")
        ttl_sql = self._generate_ttl_sql()
        index_sql = self._generate_index_sql()
        index_it_ttl_sql = self._generate_index_it_ttl_sql()
        source_sql = []
        for c in self.source_cls:
            source = c.get_source_name()
            dim_index = self.get_dim_index_code(index_sql[source])
            source_sql.append(
                f"""select  
            a.comp_id as {self.ads_columns.comp_id}
            ,com.company_abbr as {self.ads_columns.company_abbr}
            ,com.industry_name as {self.ads_columns.industry_name}
            ,a.eff_year as {self.ads_columns.eff_year}
            ,a.index_version as {self.ads_columns.index_version}
            ,a.index_value as {self.ads_columns.ttl_cnt}
            ,b.index_value as {self.ads_columns.it_cnt}
            ,'{self.get_source_alias(c)}' as {self.ads_columns.source_name}
            ,a.index_name as {self.ads_columns.index_name}
            ,c.index_value as {self.ads_columns.index_value}
            from (
                select /*+ MAPJOIN(b) */  
                    comp_id,index_code,index_name,b.index_version
                    ,index_value
                    ,eff_year from {ttl_sql[source]} a 
                join {dim_index} b ) a
            join {self._dim_com} com on a.comp_id = com.comp_id
            left join {index_it_ttl_sql[source]} b on a.comp_id = b.comp_id and a.eff_year = b.eff_year
            left join {index_sql[source]} c on a.comp_id = c.comp_id and a.eff_year = c.eff_year and a.index_code = c.index_code
            """
            )
        combined_sql = " union all ".join(source_sql)
        ads_cols = ",".join(self.ads_columns.__dict__.values())
        dashboard_sql = f"""
            {'INSERT OVERWRITE TABLE ads_digital_talent_index_dashboard ' if not test else ' ' }
            select {ads_cols} from ({combined_sql});
        """
        return dashboard_sql

    def generate_dashboard_data(self):
        dashboard_sql = self._generate_dashboard_sql()
        print(
            f"start to generate dashboard data: [{self.comp_type},'{self.index_version}','{self.index_codes}'] ..."
        )
        SqlRunner(dashboard_sql).run()
        print(f"done: dashboard data")

    def get_source_alias(self, cls: Type[TIndexGenerator]):
        _source_map = {"talent": "stock"}
        return _source_map.get(cls.get_source_name(), cls.get_source_name())

    def get_result_df(self):
        dashboard_sql = self._generate_dashboard_sql(test=True)
        result = SqlRunner(dashboard_sql).to_pandas()
        return result



from .common_index_generator import CommonIndexGenerator


class DemandIndexGenerator(CommonIndexGenerator):
    # overload the baseclass
    _sql_comp_map_lst_com = """ ( 
        SELECT  distinct full_stock_id AS comp_id
            ,tmp.source_comp_id
            ,nvl(abbr_cn,company_name_cn) company_abbr
            FROM    dim_china_listed_company a
            LATERAL VIEW EXPLODE(social_credit_codes) tmp AS source_comp_id
                WHERE pt = MAX_PT('dim_china_listed_company')
            )
        """
    __sql_src_ttl_lst_com = """ (
        select social_credit_code as source_comp_id
            ,job_id as entity_id
            ,pt as start_year
            ,pt as end_year
            ,'total' as tag_udf_version
            from dwd_scc_com_job_posting
            where pt>'2015'
            )
        """
    _TABLE_SOURCE_LST_COM = "dwd_tag_scc_com_job_posting"
    _sql_src_tag_lst_com = (
        """ (
            select social_credit_code as source_comp_id
            ,job_id as entity_id
            ,activity_year as start_year
            ,activity_year as end_year
            ,tag_udf_version
            from """
        + _TABLE_SOURCE_LST_COM
        + """
            where tag_udf_version = {tag_udf_version}
            and {tag_condition}
            )
        """
    )

    @property
    def _sql_src_ttl_lst_com(self):
        return self.__sql_src_ttl_lst_com

    @property
    def _table_source_tag_lst_com(self):
        return self._TABLE_SOURCE_LST_COM

    def get_sql_source_tag_lst_com(
        self, tag_condition: str, tag_udf_version: str
    ) -> str:
        return self._sql_src_tag_lst_com.format(
            **{
                "tag_udf_version": tag_udf_version,
                "tag_condition": tag_condition,
            }
        )

    @classmethod
    def get_source_name(cls):
        return "demand"

    @classmethod
    def get_index_suffix(cls):
        return "D"



from typing import List, Optional
from treelib import Node, Tree
from tagging_index._udf.tag_converter import TagConfig


class TagNode(Node):
    def __init__(
        self, tag=None, identifier=None, expanded=True, data: Optional[TagConfig] = None
    ):
        super().__init__(tag, identifier, expanded, data)
        self.level = data.tag_level if data else 0


class TagTree(Tree):

    def __init__(
        self,
        tag_configs: List[TagConfig],
    ):
        super().__init__(node_class=TagNode)
        tag_configs.sort(key=lambda x: x.tag_level)
        self.tag_configs = tag_configs
        if tag_configs:
            root_tag = self.create_node("**tag_root**")
            invalid_parent = self.create_node("**invalid_parent**", parent=root_tag)
            for c in tag_configs:
                p = c.parent_tags[0] if c.parent_tags else root_tag.identifier
                if self.contains(p):
                    self.create_node(c.tag, c.tag, parent=p, data=c)
                else:
                    invalid_p = self.create_node(p, p, parent=invalid_parent)
                    self.create_node(c.tag, c.tag, parent=invalid_p, data=c)
            # remove invalid tag if no child
            invalid_count = len(self.children(invalid_parent.identifier))
            if invalid_count == 0:
                self.remove_node(invalid_parent.identifier)

    def show_level(self, nid=None, levels=0):
        """print tree from nid to below levels, default levels = 0 to print all level below nid

        Args:
            levels (int, optional): how may levels under nid to print. Defaults to 0.
        """
        node = self.get_node(nid)
        if node:
            level = node.level + levels
            path = [x for x in self.rsearch(nid)]
            children = self.children(node.identifier)
            all_node = path + [
                x.identifier for x in children if x.level <= level or levels == 0
            ]
            print(self.show(filter=lambda x: x.identifier in all_node, stdout=False))
        else:
            level = levels
            print(
                self.show(
                    nid=nid,
                    filter=lambda x: (x.level <= level) or levels == 0,
                    stdout=False,
                )
            )



from odps.udf import annotate


# UDF类
@annotate("string,bigint,array<string>->array<string>")
class match_tag(object):
    def __init__(self):
        import sys

        deps = ["tag_matcher.py", "tag_converter.py"]
        for dep in deps:
            sys.path.insert(0, f"work/{dep}")
        from tag_matcher import TagMatcher
        from tag_converter import TagConverter

        self.converter = TagConverter()
        self.tag_version, self.tag_configs = self.converter.udf_res_to_config(
            "res_dim_tag_new"
        )
        self.matcher = TagMatcher()
        self.matcher.load_tags(self.tag_configs)

    def evaluate(self, content, return_type, tag_range):
        # !UDF_VERSION 会在dataworks 上发布时自动添加
        match_results = self.matcher.match_result(content, return_type, tag_range, self.tag_version, UDF_VERSION)  # type: ignore
        return match_results



from .common_index_generator import CommonIndexGenerator


class InnovationIndexGenerator(CommonIndexGenerator):
    # overload the baseclass
    _sql_comp_map_lst_com = """ ( 
        SELECT  distinct full_stock_id AS comp_id
            ,tmp.source_comp_id
            ,nvl(abbr_cn,company_name_cn) company_abbr
            FROM    dim_china_listed_company a
            LATERAL VIEW EXPLODE(social_credit_codes) tmp AS source_comp_id
                WHERE pt = MAX_PT('dim_china_listed_company')
            )
        """
    __sql_src_ttl_lst_com = """ (
        with app_no as (
            SELECT distinct application_no
            ,SUBSTR(application_date,1,4) AS application_year from dwd_patent_desc 
            WHERE   pt = MAX_PT ('dwd_patent_desc') 
            and nullif(content,'') is not null
        )
        ,entity as (
            select distinct application_no,ic_usc_code as social_credit_code
            from dwd_patent_entity 
        LATERAL VIEW EXPLODE(ic_usc_codes) tmp AS ic_usc_code 
        where pt = MAX_PT('dwd_patent_entity') and ic_usc_code <>""
        )
        select social_credit_code as source_comp_id
                    ,a.application_no as entity_id
                    ,application_year as start_year
                    ,application_year as end_year
                    ,'total' as tag_udf_version
                    from app_no a join entity b on a.application_no=b.application_no
            )
        """
    __TABLE_SOURCE_LST_COM = "dwd_tag_patent_china"
    __sql_src_tag_lst_com = (
        """ (
            select social_credit_code as source_comp_id
            ,application_no as entity_id
            ,application_year as start_year
            ,application_year as end_year
            ,tag_udf_version
            from """
        + __TABLE_SOURCE_LST_COM
        + """
            where tag_udf_version = {tag_udf_version}
            and {tag_condition}
            )
        """
    )

    @property
    def _sql_src_ttl_lst_com(self):
        return self.__sql_src_ttl_lst_com

    @property
    def _table_source_tag_lst_com(self):
        return self.__TABLE_SOURCE_LST_COM

    def get_sql_source_tag_lst_com(
        self, tag_condition: str, tag_udf_version: str
    ) -> str:
        return self.__sql_src_tag_lst_com.format(
            **{
                "tag_udf_version": tag_udf_version,
                "tag_condition": tag_condition,
            }
        )

    @classmethod
    def get_source_name(cls):
        return "innovation"

    @classmethod
    def get_index_suffix(cls):
        return "I"



import json
from typing import Optional, Tuple
from typing import Dict, List
from dataclasses import dataclass, field


@dataclass
class CurrentConfig:
    ignore: bool
    local_file: str
    odps_version: str


@dataclass
class ConfigData:
    csv: List[str]
    json: List[str]
    current_config: CurrentConfig


@dataclass
class ConfigOutput:
    log: str
    format: str
    output_file: str


@dataclass
class PreSuffixConfig:
    kw: str
    category: Optional[str] = field(default=None)
    data_phase: Optional[str] = field(default=None)
    distance: int = field(default=0)


@dataclass
class TagConfig:
    tag: str
    language: str
    tag_level: int
    parent_tags: List[str] = field(default_factory=list)
    is_keyword: Optional[bool] = field(default=False)
    prefix: List[PreSuffixConfig] = field(default_factory=list)
    suffix: List[PreSuffixConfig] = field(default_factory=list)
    category: Optional[str] = field(default=None)
    data_phase: Optional[str] = field(default=None)
    data_cycle: Optional[str] = field(default=None)


class TagConverter:
    def __init__(self):
        """convert tag config between raw csv, json file, odps version, config objects"""
        self.standardizer = TagStandardizer()

    def config_to_df(self, tag_configs: List[TagConfig]):
        """Convert the tag config list to a DataFrame. standardize format"""
        import pandas as pd

        df = pd.DataFrame(tag_configs)
        df = self.standardizer.apply_df(df)
        return df

    def df_to_json(self, dataframe, file_path):
        """Convert the DataFrame to JSON and save it to a json file."""
        # 动态导入,避免udf(不使用pandas) 引用时出错
        import pandas as pd

        df: pd.DataFrame = dataframe
        tag_dict = df.to_dict(orient="records")
        with open(file_path, "w", encoding="utf-8") as file:
            json.dump(tag_dict, file, ensure_ascii=False, indent=4)

    def config_to_json(self, tag_configs: List[TagConfig], file_path):
        df = self.config_to_df(tag_configs)
        self.df_to_json(df, file_path)

    def udf_res_to_config(self, table_res_name: str) -> Tuple[str, List[TagConfig]]:
        """convert resource table data to tag config, only for UDF env.

        Args:
            res_table (List[]): mc table resource data
            columns (List[str]): mc table resource column name list

        Returns:
            List[TagConfig]: tag config list
        """
        from odps.distcache import get_cache_table
        from odps.distcache import get_cache_tabledesc

        desc = list(get_cache_tabledesc(table_res_name))  # type: ignore
        data = list(get_cache_table(table_res_name))  # type: ignore
        columns = list(map(lambda x: x["name"], desc[2]))
        tag_version = data[0][columns.index("tag_version")] if len(data) > 0 else "NONE"

        tag_configs: List[TagConfig] = []
        for x in data:
            prefix_s = json.loads(
                x[columns.index("prefix")] if x[columns.index("prefix")] else "[]"
            )
            prefix = [
                PreSuffixConfig(x["kw"], x["category"], x["distance"]) for x in prefix_s
            ]
            suffix_s = json.loads(
                x[columns.index("suffix")] if x[columns.index("suffix")] else "[]"
            )
            suffix = [
                PreSuffixConfig(x["kw"], x["category"], x["distance"]) for x in suffix_s
            ]
            parent_tags = json.loads(
                x[columns.index("parent_tags")]
                if x[columns.index("parent_tags")]
                else "[]"
            )
            tc = TagConfig(
                x[columns.index("tag")],
                x[columns.index("language")],
                x[columns.index("tag_level")],
                parent_tags,
                x[columns.index("is_keyword")],
                prefix,
                suffix,
                x[columns.index("category")],
                x[columns.index("data_phase")],
                x[columns.index("data_cycle")],
            )
            tag_configs.append(tc)
        return tag_version, tag_configs

    def dict_to_config(self, dict: List[Dict]) -> List[TagConfig]:
        """convert Dict to tag config

        Returns:
            List[TagConfig]: tag config list
        """
        tag_configs: List[TagConfig] = []
        for x in dict:
            prefix = self._dict_to_fixConfig(x["prefix"])
            suffix = self._dict_to_fixConfig(x["suffix"])
            tc = TagConfig(
                x["tag"],
                x["language"],
                x["tag_level"],
                x["parent_tags"],
                x["is_keyword"],
                prefix,
                suffix,
                x["category"],
                x["data_phase"],
                x["data_cycle"] if x.keys().__contains__("data_cycle") else None,
            )
            tag_configs.append(tc)
        return tag_configs or []

    def version_to_config(self, version=""):
        """read certain version from opds tag dim table, if version="", use the latest version

        Args:
            version (str, optional): tag version. Defaults to "",use latest version.
        """
        df_tags = self.version_to_df(version)
        return self.df_to_config(df_tags)

    def version_to_df(self, version=""):
        """read certain version from opds tag dim table, if version="", use the latest version

        Args:
            version (str, optional): tag version. Defaults to "",use latest version.
        """
        import pandas as pd
        from tagging_index.maxcompute.sql_runner import SqlRunner

        dim_table = "dim_digitalization_tag"
        version = f'"{version}"' if version else f'max_pt("{dim_table}")'
        sql = f"select * from {dim_table}  where tag_version={version}"
        sql_runner = SqlRunner(sql)
        df_tags: pd.DataFrame = sql_runner.to_pandas()
        return df_tags

    def df_to_config(self, df_tags) -> List[TagConfig]:
        """convert dataframe to tag config

        Args:
            df_tags (pd.DataFrame): dataframe

        Returns:
            List[TagConfig]: tag config list
        """
        tag_configs: List[TagConfig] = []
        for _, row in df_tags.iterrows():
            prefix = self._dict_to_fixConfig(row["prefix"])
            suffix = self._dict_to_fixConfig(row["suffix"])
            tc = TagConfig(
                row["tag"],
                row["language"],
                len(row["parent_tags"] if row["parent_tags"] else []) + 1,
                row["parent_tags"],
                row["is_keyword"],
                prefix,
                suffix,
                row["category"],
                row["data_phase"],
            )
            tag_configs.append(tc)
        return tag_configs

    def _dict_to_fixConfig(self, dict: List[Dict]) -> List[PreSuffixConfig]:

        presuffix = [
            PreSuffixConfig(p["kw"], p["category"], p["data_phase"], p["distance"])
            for p in (dict if dict else [])
        ]
        return presuffix


class TagStandardizer:
    def __init__(self):
        pass

    @staticmethod
    def serialize(val):
        if (
            isinstance(val, list)
            and len(val) > 0
            and isinstance(val[0], PreSuffixConfig)
        ):
            return json.dumps(
                [x.__dict__ for x in val], ensure_ascii=False
            )  # Implement a to_dict method in PreSuffixConfig
        else:
            return (
                json.dumps(val, ensure_ascii=False)
                if isinstance(val, dict) or isinstance(val, list)
                else val
            )

    @staticmethod
    def standardize_value(value):
        """Standardize a single value by stripping, converting to lower case, etc."""
        if isinstance(value, str):
            return value.strip()
        return value

    def df_serialize(self, dataframe):
        for column in dataframe.columns:
            dataframe[column] = dataframe[column].apply(self.serialize)
        return dataframe

    def apply_df(self, dataframe):
        """Apply standardization to each column in the dataframe."""
        for column in dataframe.columns:
            dataframe[column] = dataframe[column].apply(self.standardize_value)
        return dataframe

    def apply_config(self, tag_config: List[TagConfig]):
        """Apply standardization to each tag,prefix.kw,suffix.kw in the list."""
        for t in tag_config:
            t.tag = self.standardize_value(t.tag)
            for p in t.prefix or []:
                p.kw = self.standardize_value(p.kw)
            for s in t.suffix or []:
                s.kw = self.standardize_value(s.kw)
            t.parent_tags = (
                [self.standardize_value(p) for p in t.parent_tags]
                if t.parent_tags
                else []
            )

        return tag_config



import os
from setuptools import setup, find_namespace_packages

name = "tagging_index"
description = "content tagging and index generator with maxcompute"
url = "https://github.com/Digital-Transformation-Research-Center/tagging-index"  # 项目网址
# # 依赖包列表
install_requires = [
    "alibabacloud_tea_openapi>=0.3.8, <1.0.0",
    "pandas>=1.3.5",
    "pyodps>=0.11.6",
    "treelib>=1.7.0",
]

author = "t2wei"  # 作者名
author_email = "t2wei@me.com"  # 作者邮箱

# git action workflow yml文件中定义的版本号,来自于tag
version = os.getenv("PACKAGE_VERSION", "0.0.1-alpha")
# 自动发现所有包
packages = find_namespace_packages(".", include=[name + "*"], exclude=["*schemas*"])
print(f"packages:{packages}")
if not version:
    raise Exception("no version info found!")
print(f"version: {version};")
setup(
    name=name,  # 包的名称，应与项目目录名一致，且符合PyPI的要求
    version=version,  # 版本号
    author=author,
    author_email=author_email,
    description=description,
    long_description=open("README.md").read(),  # 详细描述，通常从README文件读取
    long_description_content_type="text/markdown",  # 如果README是Markdown格式
    url=url,
    packages=packages,
    include_package_data=True,
    classifiers=[  # 分类信息，帮助用户在PyPI上找到你的包
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=install_requires,
    python_requires=">=3.7",  # 指定Python版本要求
)



from .common_index_generator import CommonIndexGenerator


class TalentIndexGenerator(CommonIndexGenerator):
    # define the sql template as static property

    __sql_src_ttl_lst_com = """(
            select full_stock_id as source_comp_id
            ,user_id as entity_id
            ,start_year
            ,end_year
            ,'total' as tag_udf_version
            from dwd_cn_lst_com_pos_linkedin
            where end_year>2000
            )
        """
    __TABLE_SOURCE_LST_COM = "dwd_tag_lst_com_linkedin_position_skill"
    _sql_src_tag_lst_com = (
        """(
            select full_stock_id as source_comp_id
            ,user_id as entity_id
            ,hire_year as start_year
            ,dismiss_year as end_year
            ,tag_udf_version
            from """
        + __TABLE_SOURCE_LST_COM
        + """
            where tag_udf_version = {tag_udf_version}
            and {tag_condition}
            )
        """
    )

    @property
    def _sql_src_ttl_lst_com(self):
        return self.__sql_src_ttl_lst_com

    @property
    def _table_source_tag_lst_com(self):
        return self.__TABLE_SOURCE_LST_COM

    def get_sql_source_tag_lst_com(
        self, tag_condition: str, tag_udf_version: str
    ) -> str:
        return self._sql_src_tag_lst_com.format(
            **{
                "tag_udf_version": tag_udf_version,
                "tag_condition": tag_condition,
            }
        )

    @classmethod
    def get_source_name(cls):
        return "talent"

    @classmethod
    def get_index_suffix(cls):
        return "T"



from abc import ABC, abstractmethod
from dataclasses import dataclass
import json
from os import path
from pathlib import Path
from typing import Dict, List, TypeVar
import pandas as pd
from tagging_index.maxcompute import SqlRunner


sql_create_table = """
    CREATE TABLE IF NOT EXISTS dwd_digital_index(
	comp_id STRING,
    company_abbr string,
    index_name string,
	index_value string,
    eff_year STRING
    ) 
    PARTITIONED  by  (
    index_version string,
    comp_type string,
    index_code string
    );
    """


@dataclass(frozen=True)
class IndexSqlTemplate:
    sql_comp_map: str
    sql_source_tag_data: str
    sql_year_sequence: str
    index_version: str
    comp_type: str
    index_code: str
    index_name: str
    dwd_table: str
    # TODO add year range constraint for source data
    INDEX_TEMPLATE_SELECT = """ 
        SELECT /*+ MAPJOIN(y) */
            a.comp_id
            ,a.company_abbr
            ,"{index_name}" as index_name
            ,count(DISTINCT entity_id) as index_value
            ,y.eff_year
            ,CONCAT_ws('-','{index_version}',b.tag_udf_version) as index_version 
            ,"{comp_type}" as comp_type
            ,"{index_code}" as index_code
        FROM
        {sql_comp_map} a join {sql_source_tag_data} b on a.source_comp_id = b.source_comp_id
        join {sql_year_sequence} y 
        on b.end_year>=y.eff_year and b.start_year<=y.eff_year  
        group by a.comp_id,a.company_abbr,y.eff_year,b.tag_udf_version
        ;
        """
    INDEX_TEMPLATE_INSERT = (
        "INSERT OVERWRITE TABLE {dwd_table} PARTITION (index_version,comp_type,index_code) "
        + INDEX_TEMPLATE_SELECT
    )


class CommonIndexGenerator(ABC):
    # define the sql template as static property
    DWD_TABLE = "dwd_digital_index"
    _sql_comp_map_lst_com = """ ( 
                SELECT distinct full_stock_id as comp_id
                ,full_stock_id source_comp_id
                ,nvl(abbr_cn,company_name_cn) company_abbr
                FROM dim_china_listed_company 
                WHERE pt = MAX_PT('dim_china_listed_company')
            )
        """

    @property
    def sql_comp_map_lst_com(self):
        return self._sql_comp_map_lst_com

    @property
    def sql_comp_map(self):
        if self.comp_type == "lst_com":
            cte_comp_map = self.sql_comp_map_lst_com
        else:
            raise ValueError(f"not implemented for comp_type: {self.comp_type}")
        return cte_comp_map

    @property
    @abstractmethod
    def _sql_src_ttl_lst_com(self) -> str:
        pass

    @property
    @abstractmethod
    def _table_source_tag_lst_com(self) -> str:
        pass

    @abstractmethod
    def get_sql_source_tag_lst_com(
        self, tag_condition: str, tag_udf_version: str
    ) -> str:
        pass

    _sql_year_sequence = """ 
            (SELECT EXPLODE(SEQUENCE({start_year},{end_year})) as (eff_year))
    """

    @property
    def sql_year_sequence(self):
        sql_year_sequence = self._sql_year_sequence.format(
            **{"start_year": self.start_year, "end_year": self.end_year}
        )
        return sql_year_sequence

    @classmethod
    @abstractmethod
    def get_source_name(cls) -> str:
        pass

    @classmethod
    @abstractmethod
    def get_index_suffix(cls) -> str:
        pass

    def __init__(self, tag_desc_file: str):
        self._tag_desc_file = tag_desc_file
        self.tag_udf_version = ""
        self.start_year = 2000
        self.end_year = 2024
        self.comp_type = "lst_com"
        meta_json = json.load(open(self._tag_desc_file))
        self.version = meta_json["version"]
        self.tag_version = meta_json["tag_version"]
        self.index_meta = {
            index_definition["index_code"]: index_definition
            for index_definition in meta_json["index_definition"]
        }
        self.global_exclude = meta_json.get("exclude", {})
        # self.df_index_meta=pd.DataFrame(self.index_meta)

    @property
    def index_codes(self) -> List[str]:
        return [
            f"{index_code}_{self.get_index_suffix()}"
            for index_code in list(self.index_meta.keys())
        ]

    def generate_sql(self, index_codes: List[str] = []) -> Dict[str, str]:
        return self._generate_sql(index_codes)

    def __get_source_version_cond(self, source_ver: str) -> str:
        src_data_ver_cond = (
            f'"{source_ver}"'
            if source_ver
            else f'max_pt("{self._table_source_tag_lst_com}")'
        )
        return src_data_ver_cond

    def _generate_sql(
        self, index_codes: List[str] = [], return_result=False
    ) -> Dict[str, str]:
        index_generator_sql = {}
        if index_codes is None or len(index_codes) == 0:
            index_codes = list(self.index_meta.keys())

        for index_code in index_codes:
            template = self.get_index_sql_template(index_code)
            sql_template = (
                template.INDEX_TEMPLATE_SELECT
                if return_result
                else template.INDEX_TEMPLATE_INSERT
            )
            index_generator_sql[index_code] = sql_template.format(**template.__dict__)

        return index_generator_sql

    def get_index_sql_template(self, index_code: str) -> IndexSqlTemplate:
        src_data_ver_cond = self.__get_source_version_cond(self.tag_udf_version)
        tag_condition = self._generate_tag_condition(
            self.index_meta[index_code]["condition"], self.tag_version
        )
        if self.comp_type == "lst_com":
            sql_source_tag_data = self.get_sql_source_tag_lst_com(
                tag_condition, src_data_ver_cond
            )
        else:
            raise ValueError(f"not implemented for comp_type: {self.comp_type}")
        index_sql_template = IndexSqlTemplate(
            sql_comp_map=self.sql_comp_map,
            sql_source_tag_data=sql_source_tag_data,
            sql_year_sequence=self.sql_year_sequence,
            index_version=self.version,
            comp_type=self.comp_type,
            index_code=f"{index_code}_{self.get_index_suffix()}",
            index_name=self.index_meta[index_code]["index_name"],
            dwd_table=self.DWD_TABLE,
        )
        return index_sql_template

    def generate_index(self, index_codes: List[str] = []):
        index_sql = self.generate_sql(index_codes)
        print(
            f"*****{self.get_source_name()} index generater config: [comp:{self.comp_type},tag:{self.tag_version},source:{self.tag_udf_version or 'latest source version'}] ..."
        )
        for index_code, sql in index_sql.items():
            index_name = self.index_meta[index_code]["index_name"]
            print(f"generating index: {index_name}[code:{index_code}] ...")
            result = self._run(sql)
            print(
                f"done: {index_name}[{index_code}](total records:{result[0].outputs[0].rows})"
            )

    def generate_index_ttl(self):
        index_sql = self.generate_total_sql()
        print(
            f"generating total count index: [{self.get_source_name()},{self.comp_type}] ..."
        )
        result = self._run(index_sql)
        print(f"done: total count index (total records:{result[0].outputs[0].rows})")

    def generate_total_sql(self):
        index_config = IndexSqlTemplate(
            sql_comp_map=self.sql_comp_map,
            sql_source_tag_data=self._sql_src_ttl_lst_com,
            sql_year_sequence=self.sql_year_sequence,
            index_version=self.get_source_name(),
            comp_type=self.comp_type,
            index_code="ttl_cnt",
            index_name="total count of source",
            dwd_table=self.DWD_TABLE,
        )
        sql = index_config.INDEX_TEMPLATE_INSERT.format(**index_config.__dict__)
        return sql

    def _generate_tag_condition(self, condition: dict, tag_version):
        category_sql_parts = []
        dim_table = "dim_digitalization_tag"
        dim_view = "v_dim_digital_tag_level"
        or_cond = condition.get("or", {})
        exclude = condition.get("exclude", {}) or self.global_exclude
        tag_version_cond = (
            f'"{tag_version}"' if tag_version else f'max_pt("{dim_table}")'
        )
        exclude_tag_cond = (
            " tag not in ("
            + ",".join([f"'{tag}'" for tag in exclude.get("tags")])
            + ")"
            if exclude.get("tags", [])
            else "1=1"
        )
        for level, tags in or_cond.items():
            tags_str = ", ".join([f"'{tag}'" for tag in tags])
            category_sql_parts.append(f"tag_lv{level[-1]} in ({tags_str})")
        if not category_sql_parts:
            category_sql_parts = ["1=1"]
        category_sql_query = f""" tag in (SELECT tag FROM {dim_view} 
            WHERE tag_version={tag_version_cond}
            AND ( {' OR '.join(category_sql_parts)})
            and {exclude_tag_cond}
            )
        """
        return category_sql_query

    def get_index_data(self, index_code: str) -> pd.DataFrame:
        """return specified index data in dataframe rather than insert into table

        Args:
            index_code (str)

        Returns:
            pd.DataFrame:
        """
        sql = self._generate_sql([index_code], return_result=True).get(index_code)
        if sql:
            df = SqlRunner(sql).to_pandas()
            return df
        else:
            raise ValueError(f"index code {index_code} not found")

    def _run(self, sql):
        sql_runner = SqlRunner(sql)
        sql_runner.run()
        summary = sql_runner.get_summary()
        # print(summary[0].outputs_partitions)
        return summary

    @staticmethod
    def get_index_schema():
        schema_file = Path(__file__).parent / "index_tag_schema.json"
        return json.load(open(schema_file, encoding="utf-8", mode="r"))


TIndexGenerator = TypeVar("TIndexGenerator", bound=CommonIndexGenerator)



import json
import os
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Optional

from tagging_index._udf.tag_matcher import TagMatcher
from tagging_index._udf.tag_converter import TagConverter
from .initialize_maxcompute import _maxcompute_setup as maxcompute_setup

# 文件和版本变量
RES_FILES = {
    "udf_match_tag": "UDF_MATCH_TAG_VERSION",
    "tag_converter": "TAG_CONVERTER_VERSION",
    "tag_matcher": "TAG_MATCHER_VERSION",
}
MAIN_FILE = "udf_match_tag"
DIM_RES_NAME = "res_dim_tag_new"


class UdfRelease:
    def __init__(self):
        self.o = maxcompute_setup().default_odps

    def release_udf(self):
        self._update_py_res()
        updated_files = self._check_res_update()
        self.version = self._update_main_file(updated_files)
        return self.version

    def _read_file_and_version(self, filename, version_info=True):
        if not self.o.exist_resource(filename):
            return "", 0
        with self.o.open_resource(filename, mode="r") as fp:
            content = fp.read()
            version_line = content.split("\n")[0]
            version = int(version_line.split("=")[1]) if version_info else 0
            return (content if not version_info else content.split("\n", 1)[1], version)

    def _check_res_update(self):
        updated_files = {}
        for filename, version_var in RES_FILES.items():
            original_content, _ = self._read_file_and_version(
                f"{filename}.py", version_info=False
            )
            release_file = f"{filename}_release.py"
            release_content, release_version = self._read_file_and_version(release_file)

            # 检查发布文件是否存在，以及内容是否有更新
            if (
                not self.o.exist_resource(release_file)
                or original_content != release_content
            ):
                new_version = release_version + 1 if release_version else 1
                updated_content = f"{version_var}={new_version}\n" + original_content
                updated_files[filename] = {
                    "content": updated_content,
                    "version": new_version,
                    "updated": True,
                }
                # 更新版本文件和发布文件
                versioned_filename = f"{filename}_v{new_version}.py"
                self.o.create_resource(
                    versioned_filename, "py", file_obj=updated_content
                )
                if self.o.exist_resource(release_file):
                    self.o.delete_resource(release_file)
                self.o.create_resource(release_file, "py", file_obj=updated_content)
                print(
                    f"...Updated {filename} to version {new_version} and refreshed {release_file}"
                )
            else:
                updated_files[filename] = {
                    "content": original_content,
                    "version": release_version,
                    "updated": False,
                }
        return updated_files

    def _update_main_file(self, updated_files: Dict[str, Any]):
        main_file = f"{MAIN_FILE}_main.py"  # 发布文件名
        res_versions = ";".join(
            [f"{k}:{v['version']}" for k, v in updated_files.items()]
        )
        main_version = "@".join([str(v["version"]) for k, v in updated_files.items()])
        udf_version = main_version
        if any(f["updated"] for f in updated_files.values()):
            # 构建包含所有资源文件版本号的字符串
            version_lines = f"UDF_VERSION='{udf_version}' # {res_versions} \n"
            print(version_lines)
            main_content = updated_files[MAIN_FILE]["content"]
            # 更新主文件内容
            new_release_content = version_lines + "\n" + main_content
            if self.o.exist_resource(main_file):
                self.o.delete_resource(main_file)
            self.o.create_resource(
                main_file,
                "py",
                file_obj=new_release_content,
                comment=f"Updated on {datetime.now().strftime('%Y-%m-%d %H:%M:%S')}",
            )
            print(
                f"...Updated {main_file} with new UDF version({udf_version}) and resources' versions."
            )
        else:
            print(
                f"...No significant updates detected; main file update skipped. current version: {main_version}"
            )
        return udf_version

    def create_udf_function(self, updated_files: Dict[str, Any]):
        if not self.o.exist_function("udf_match_tag"):
            resources = [
                f"{f if f!=MAIN_FILE else f'{MAIN_FILE}_main'}.py"
                for f in updated_files.keys()
            ]
            resources.append(DIM_RES_NAME)
            self.o.create_function(
                "udf_match_tag",
                class_type=f"{MAIN_FILE}_main.match_tag",
                resources=resources,
            )

    def update_dim_resource(
        self, dim_table="dim_digitalization_tag", version: str = ""
    ):
        """update dim resouce to assigned version, or use latest version if version is ""
        Args:
            version (str, optional): udf version. Defaults to "".
        """

        version_cond = f'"{version}"' if version else f'max_pt("{dim_table}")'
        res_table = "dim_digital_tag_resource_new"
        sql_drop_res_table = f"drop table if exists {res_table};"
        sql_res_create = f"""
            create table {res_table} as 
            -- 不支持复杂类型, 不支持分区
            select tag,
            tag_level,
            to_json(prefix) prefix,
            is_keyword,
            to_json(suffix) suffix,
            language,
            category,
            data_cycle,
            data_phase,
            to_json(parent_tags) parent_tags,
            tag_version from {dim_table} where tag_version={version_cond} ;
        """
        self.o.execute_sql(sql_drop_res_table, hints={"odps.sql.submit.mode": "script"})
        self.o.execute_sql(sql_res_create, hints={"odps.sql.submit.mode": "script"})
        res = self.o.get_resource(DIM_RES_NAME)
        res.update(table_name=res_table)
        print(f"...{res_table}->{DIM_RES_NAME} updated")

    def _update_py_res(self):
        # 本地上传udf相关py资源
        udf_res_dir = f"{Path(__file__).resolve().parent.parent}/udf"
        for filename, version_var in RES_FILES.items():
            source_file = os.path.join(udf_res_dir, f"{filename}.py")
            if os.path.exists(source_file):  # 确认文件存在于当前目录
                print(f"...updating resource {filename}.py")
                if self.o.exist_resource(f"{filename}.py"):
                    self.o.delete_resource(f"{filename}.py")
                self.o.create_resource(
                    f"{filename}.py", "py", fileobj=open(source_file, "rb")
                )
            else:
                raise FileNotFoundError(
                    f"File {filename} does not exist in the udf directory."
                )

    def test_udf(
        self,
        contents: List[str],
        return_type,
        tag_range: Optional[List[str]] = None,
        alias_dim_table="",
    ):
        """use assigned dim_version to test udf, if version="", use the current dim resource

        Args:
            alias_dim_table (str, optional): tag dim table for alias resource. Defaults to "" to use current resource
        """

        if tag_range is not None:
            tag_range_list = ",".join(f"'{t}'" for t in tag_range)
            tag_range_param = f"Array({tag_range_list})"
        else:
            tag_range_param = "null"
        content_values = [f'("{c}")' for c in contents]
        content_table = f"select * from values {','.join(content_values)} as a(content)"
        sql_udf_test = f"""select content
            ,udf_match_tag(content,{return_type},{tag_range_param}) as result 
            from ({content_table})"""
        if alias_dim_table:
            alias_name = "res_dim_tag_alias"
            if self.o.exist_resource(alias_name):
                table_resource = self.o.get_resource(alias_name)
                table_resource.update(table_name=alias_dim_table)
            else:
                self.o.create_resource(
                    alias_name, "table", table_name=alias_dim_table, temp=True
                )
            sql_instantance = self.o.execute_sql(
                sql_udf_test,
                aliases={DIM_RES_NAME: alias_name},
                hints={"odps.sql.python.version": "cp37"},
            )
        else:
            sql_instantance = self.o.execute_sql(
                sql_udf_test,
                hints={"odps.sql.python.version": "cp37"},
            )
        df = sql_instantance.open_reader().to_pandas()
        # result = df.to_dict(orient="records")
        return df

    def test_local(
        self,
        contents: List[str],
        return_type,
        tag_range: List[str] = [],
        dim_tag_config="",
    ):
        current_file = Path(__file__).resolve()
        config_file = (
            f"{current_file.parent.parent}/tag_data/output/new_config_validated.json"
            if dim_tag_config == ""
            else dim_tag_config
        )
        conv = TagConverter()
        tag_config = conv.dict_to_config(json.load(open(config_file)))
        matcher = TagMatcher()
        matcher.load_tags(tag_config)

        match_results = {
            c: matcher.match_result(c, return_type, tag_range) for c in contents
        }
        return match_results
